@@ -140,4 +140,8 @@ int hname_translate(js_string *hostname, int qtype);
 
 int email_translate(js_string *hostname);
 
+int make_question(q_question *question, js_string *js);
+int read_question(js_string *js, q_question *question, int offset);
+int read_ns(js_string *in, js_string *out, int offset);
+int read_txt(js_string *in, js_string *out, int offset);
 
