@@ -145,6 +145,8 @@ int mhash_put_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
     data->next = data->ip = 0;
     data->query = data->data = 0;
     data->seen = 0;
+    data->zap = 0;
+    data->ptr = 0;
 
     /* Store the simple data in the rr data */
     data->expire = 0;
@@ -224,6 +226,8 @@ int mhash_add_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
     data->next = data->ip = 0;
     data->query = data->data = 0;
     data->seen = 0;
+    data->zap = 0;
+    data->ptr = 0;
 
     /* Store the simple data in the rr data */
     data->expire = 0;
@@ -292,6 +296,68 @@ int mhash_add_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
 
     }
 
+/* Add a PTR pointer to the first record in a given chain of records
+   (if it does not already have a PTR record)
+   This is used by the recursive code, hence we need to be careful that
+   we don't make the data inconsistant
+   Input: Pointer to hash, host name, pointer to string with value
+          of PTR record
+   Output: JS_ERROR on error, JS_SUCCESS on success
+*/
+
+int mhash_add_ptr(mhash *hash, js_string *query, js_string *value) {
+
+    rr *point;
+    js_string *new;
+    mhash_e spot_data;
+
+    /* "NotFina" means that we do not add this entry to the cache */
+    if(value->unit_count == 7 && strncmp(value->string,"NotFina",8) == 0) {
+        return JS_ERROR;
+	}
+
+    /* Create a js_string object to store the raw ptr answer */
+    if((new=js_create(value->unit_count + 1,value->unit_size)) == 0) {
+        /* Jaakko things that we should add a js_destroy(new); here.
+           I disagree--if the creation of the object returns a 0,
+           it was never created. */
+        return JS_ERROR;
+        }
+    if(js_copy(value,new) == JS_ERROR) {
+        js_destroy(new);
+        return JS_ERROR;
+        }
+    /* And put a pointer to that js_string in the rr data */
+
+    /* Find the element in the big hash to add this data to */
+    spot_data = mhash_get(hash,query);
+    if(spot_data.value == 0) { /* If the element does not exist */
+        js_destroy(new);
+        return JS_ERROR; /* Or maybe we should just add the data... */
+        /* ret = mhash_put(hash,query,data,MARA_DNSRR); (Error check) */
+        }
+    /* Exit if the data pointed to is *not* a DNS resource record */
+    if(spot_data.datatype != MARA_DNSRR) {
+        js_destroy(new);
+        return JS_ERROR;
+        }
+
+    /* OK, now all we have to do is add the ptr if it is not already set */
+
+    /* Point to the first link in the chain */
+    point = spot_data.value;
+
+    /* In order to avoid inconsistant data, we only add the PTR once */
+    if(point->ptr != 0) {
+        js_destroy(new);
+        return JS_SUCCESS; /* It was, technically, sucessful */
+        }
+
+    point->ptr = new;
+    return JS_SUCCESS;
+
+    }
+
 /* Add a pointer to an IP to the first record in a given chain of records
    (if it does not already have an IP)
    This is used by the recursive code, hence we need to be careful that
@@ -319,6 +385,7 @@ int mhash_add_ip(mhash *hash, js_string *query, js_string *value) {
     data->query = data->data = 0;
     data->seen = 0;
     data->zap = 0;
+    data->ptr = 0;
 
     /* Since we are adding an IP, the rrtype is always RR_A */
     rrtype = RR_A;
@@ -327,9 +394,9 @@ int mhash_add_ip(mhash *hash, js_string *query, js_string *value) {
     /* Create a js_string object to store the raw binary answer */
     if((new=js_create(value->unit_count + 1,value->unit_size)) == 0) {
         js_dealloc(data);
-	/* Jaakko things that we should add a js_destroy(new); here.
-	   I disagree--if the creation of the object returns a 0, 
-	   it was never created. */
+        /* Jaakko things that we should add a js_destroy(new); here.
+           I disagree--if the creation of the object returns a 0,
+           it was never created. */
         return JS_ERROR;
         }
     if(js_copy(value,new) == JS_ERROR) {
@@ -374,10 +441,10 @@ int mhash_add_ip(mhash *hash, js_string *query, js_string *value) {
     if((new_query = js_create(point->data->unit_count + 3,1)) == 0) {
         js_dealloc(data);
         js_destroy(new);
-	/* Jaakko feels that we should have a js_destroy(new_query); here.
-	   I disagree, since the memory was never allocated, and it is
-	   possible that a free(0) will cause a segfault--something we
-	   definitly don't want on a daemon */
+        /* Jaakko feels that we should have a js_destroy(new_query); here.
+           I disagree, since the memory was never allocated, and it is
+           possible that a free(0) will cause a segfault--something we
+           definitly don't want on a daemon */
         return JS_ERROR;
         }
     if(js_copy(point->data,new_query) == JS_ERROR) {
@@ -499,6 +566,7 @@ int populate_main(mhash *maintable, js_string *error) {
     js_file *desc;
     uint32 ttl;
     int rr_type, linenum = 0, rrnum = 0, in_ns = 1;
+    int result;
 
     /* Sanity check */
     if(js_has_sanity(error) == JS_ERROR)
@@ -513,11 +581,11 @@ int populate_main(mhash *maintable, js_string *error) {
         return JS_ERROR;
     if((filename = js_create(256,1)) == 0)
         return JS_ERROR;
-    if((line = js_create(256,1)) == 0)
+    if((line = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
         return JS_ERROR;
     if(js_set_encode(line,MARA_LOCALE) == JS_ERROR)
         return JS_ERROR;
-    if((pline = js_create(256,1)) == 0)
+    if((pline = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
         return JS_ERROR;
     if(js_set_encode(pline,MARA_LOCALE) == JS_ERROR)
         return JS_ERROR;
@@ -527,7 +595,7 @@ int populate_main(mhash *maintable, js_string *error) {
         return JS_ERROR;
     if((query = js_create(256,1)) == 0)
         return JS_ERROR;
-    if((data = js_create(256,1)) == 0)
+    if((data = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
         return JS_ERROR;
     if((zone = js_create(MAX_ZONE_SIZE,1)) == 0)
         return JS_ERROR;
@@ -540,18 +608,18 @@ int populate_main(mhash *maintable, js_string *error) {
     csvs = (mhash *)dvar_raw(0);
     if(csvs == 0) {
         /* A csv1 hash is no longer mandatory, now that MaraDNS has recursive
-	   capability */
+           capability */
         return 0;
         /*js_qstr2js(error,L_INITCSV); */
-	/*"csv1 hash is not correctly initialized.\nMake sure to have csv1 initialized with csv1 = {}, and that it has one or\nmore elements." */
+        /*"csv1 hash is not correctly initialized.\nMake sure to have csv1 initialized with csv1 = {}, and that it has one or\nmore elements." */
         /*return -2;*/
         }
     /* Point to the first key in the hash */
     if(mhash_firstkey(csvs,zone) == 0) {
-        /* We actually do not need to have anything in the hash, in 
-	   the case of people running a recursive nameserver */
+        /* We actually do not need to have anything in the hash, in
+           the case of people running a recursive nameserver */
         return 0;
-	}
+        }
     do {
     /*while((offset = mhash_nextkey(csvs,offset)) != 0) */
         /* If this hash element points to live data */
@@ -596,27 +664,20 @@ int populate_main(mhash *maintable, js_string *error) {
             while(!js_buf_eof(desc)) {
 
                 /* Get the line */
-                if(js_buf_getline(desc,line) == JS_ERROR)
+                result = js_buf_getline(desc,line);
+                if(result == JS_ERROR)
                     return JS_ERROR;
+                else if(result == -2) {
+                    printf("%s%d%s",L_LINE_NUMBER,linenum + 1,L_TOO_LONG);
+                    show_esc_stdout(filename);
+                    printf("\n");
+                    /* "line number ... is too long in file " */
+                    return JS_ERROR;
+                    }
 
                 /* Increment the Line Number */
                 linenum++;
 
-                /* Handle blank zone files and zone files with only
-                   SOA and authoritative NS records */
-                if(js_buf_eof(desc) && in_ns && line->unit_count == 0) {
-                    /* Yes, there is the special case of a zone with
-                       no records */
-                    if(js_length(soaline) == 0)
-                        continue;
-                    /* Add the SOA in the special case of the zone
-                       only having SOA and NS records */
-                    if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA)
-                        return JS_ERROR;
-                    add_rr_to_bighash(maintable,query,data,ttl,udpzone);
-                    break;
-                    }
-
                 /* We stop processing this file if we are at the end */
                 if(js_buf_eof(desc) && line->unit_count == 0)
                     break;
@@ -667,6 +728,7 @@ int populate_main(mhash *maintable, js_string *error) {
                        record knows what NS records to place in the NS
                        section of the answer) */
                     else if(rrnum == 1) {
+                        in_ns = 1;
                         if(js_copy(pline,soaline) == JS_ERROR)
                             return JS_ERROR;
                         }
@@ -715,6 +777,20 @@ int populate_main(mhash *maintable, js_string *error) {
                         add_rr_to_bighash(maintable,query,data,ttl,udpzone);
                     }
                 }
+            /* Handle blank zone files and zone files with only
+               SOA and authoritative NS records */
+            if(in_ns) {
+                /* Yes, there is the special case of a zone with
+                   no records */
+                if(js_length(soaline) == 0)
+                    continue;
+                /* Add the SOA in the special case of the zone
+                   only having SOA and NS records */
+                if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA)
+                    return JS_ERROR;
+                add_rr_to_bighash(maintable,query,data,ttl,udpzone);
+                }
+
             js_close(desc);
             }
         } while(mhash_nextkey(csvs,zone) != 0);
@@ -800,14 +876,21 @@ int make_ddip(mhash *bighash, js_string *query) {
         return JS_ERROR;
         }
 
-    return mhash_put_rr(bighash,query,js_ip,8675309,1,0);
+    if (mhash_put_rr(bighash,query,js_ip,8675309,1,0) == JS_ERROR) {
+        js_destroy(js_ip);
+        return JS_ERROR;
+        }
+    else {
+        js_destroy(js_ip);
+        return JS_SUCCESS;
+        }
 
     }
 
 /* Make pointers in MX, NS, and CNAME records so we can have A records in the
    AR section (AN section with CNAMEs)
    Input: A pointer to the hash to populate, a pointer to the string to
-          put an error message input an error message in
+          put an error message in
    Output: JS_ERROR on error, -2 on non-fatal error, JS_SUCCESS on success
  */
 
@@ -822,13 +905,16 @@ int add_an(mhash *bighash, js_string *error) {
         return JS_ERROR;
     if((query = js_create(MAX_ZONE_SIZE,1)) == 0) {
         js_destroy(a_query);
-	return JS_ERROR;
-	}
+        return JS_ERROR;
+        }
 
     /* Go through the big hash, looking for MX, NS, and CNAME records */
 
-    if(mhash_firstkey(bighash,query) <= 0)
-        return 0;
+    if(mhash_firstkey(bighash,query) <= 0) {
+	js_destroy(a_query);
+	js_destroy(query);
+	return 0;
+	}
 
     do {
         /* Determine if this is a MX, NS, or CNAME query */
@@ -868,6 +954,8 @@ int add_an(mhash *bighash, js_string *error) {
                 }
             }
         } while(mhash_nextkey(bighash,query) != 0);
+    js_destroy(a_query);
+    js_destroy(query);
     return JS_SUCCESS;
     }
 
