@@ -31,6 +31,7 @@
 #include "../dns/functions_dns.h"
 #include "../parse/functions_parse.h"
 extern int decomp_init();
+extern int show_timestamp();
 
 /* Our one global variable: The number of children processes running */
 int num_children = 0;
@@ -310,8 +311,13 @@ int serve_zone(int connect) {
         goto clean_ndzqrs;
 
     /* Open up the zone file */
-    if(js_open_read(filename,&desc) == JS_ERROR) 
+    if(js_open_read(filename,&desc) == JS_ERROR) {
+        show_timestamp();
+        printf("Unable to open zone file ");
+        show_esc_stdout(filename);
+	printf("\n");
         goto clean_ndzqrs;
+	}
 
     /* Build up the header, which is the same for all the replies */ 
     header.qr = header.aa = 1;
@@ -342,14 +348,25 @@ int serve_zone(int connect) {
             if(js_buf_getline(&desc,data) <= JS_ERROR)
                 goto clean_ndzqrs;
             /* Process the % character and any \ escape sequences */
-            if(bs_process(data,name,zone) == JS_ERROR)
+            if(bs_process(data,name,zone) == JS_ERROR) {
+                show_timestamp();
+                printf("Fatal syntax error in file ");
+                show_esc_stdout(filename);
+	        printf("\n");
                 goto clean_ndzqrs;
+		}
             /* Get the data from the line in the zone file */
             rr_type = parse_csv1_line(name,query,data,&ttl);
-            if(rr_type == JS_ERROR)
+            if(rr_type == JS_ERROR) {
+                show_timestamp();
+                printf("Fatal syntax error in file ");
+                show_esc_stdout(filename);
+	        printf("\n");
                 goto clean_ndzqrs;
-            if(rr_type == -2) /* Syntax error */
+		}
+            if(rr_type == -2) /* Non-fatal Syntax error */ {
                 continue;
+		}
             if(rr_type > 0 && rr_type < 65536) { /* Line with RR */
 	        /* If we have added all of the NS records, add no more */
 	        if(soa_q == 2 && rr_type != RR_NS)
@@ -764,6 +781,8 @@ on %s\nFor usage information, type in 'man maradns'" */
         /* Hackish way to clean up child processes without needlessly slowing 
            the program */
         while(waitpid(0,NULL,WNOHANG) > 0) num_children--; 
+	close(connection); /* Otherwise we can only serve MAXFD zones
+                              before it blows up */
         }
     
     }
